/*
 * deviced
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <libkmod.h>
#include <usbg/usbg.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mount.h>
#include <unistd.h>
#include <libgdbus/dbus-system.h>

#include "core/log.h"
#include "core/config-parser.h"
#include "core/device-idler.h"
#include "core/device-notifier.h"
#include "core/devices.h"
#include "core/list.h"
#include "shared/deviced-systemd.h"

#define FFS_PATH "/run/usb-host-test-ffs"
#define GADGET_SCHEME_PATH "/etc/deviced/usb-host-test/test_gadget.gs"
#define FFS_INSTANCE_NAME "usb-host-test"
#define GADGET_NAME "g1"
#define SYSTEMD_UNIT_NAME "usb-host-test.socket"
#define SYSTEMD_SERVICE_NAME "usb-host-ffs-test-daemon.service"
#define UDC_NAME "dummy_udc.0"
#define CONFIGFS_PATH "/sys/kernel/config/"

#define SYSTEMD_DBUS_PATH "/org/freedesktop/systemd1"
#define JOB_MANAGER_INTERFACE "org.freedesktop.systemd1.Manager"
#define JOB_REMOVED "JobRemoved"

static int using_legacy = 0;
static guint sigid_start;
static guint sigid_stop;
static enum device_ops_status status = DEVICE_OPS_STATUS_UNINIT;

static int load_module(const char *name, const char *options)
{
	struct kmod_ctx *ctx;
	struct kmod_module *mod;
	const char *config = NULL;
	struct kmod_list *l, *list = NULL;
	int ret = 0;
	int n = 0;

	ctx = kmod_new(NULL, &config);
	if (!ctx)
		return -1;

	ret = kmod_module_new_from_lookup(ctx, name, &list);
	if (ret < 0) {
		_E("Module %s lookup error", name);
		goto out;
	}

	kmod_list_foreach(l, list) {
		mod = kmod_module_get_module(l);
		if (!mod) {
			_E("Module %s load error", name);
			ret = -1;
			goto out;
		}

		ret = kmod_module_get_initstate(mod);
		if (ret >= 0) {
			_I("module %s already loaded", name);
			goto out; /* already loaded */
		}

		ret = kmod_module_insert_module(mod, 0, options);
		if (ret < 0) {
			_E("Module %s insert error", name);
			goto out;
		}

		++n;
	}

	if (n == 0) {
		_E("Module %s not found", name);
		ret = -1;
		goto out;
	}

	_I("Module %s loaded\n", name);
	ret = 0;

out:
	kmod_module_unref_list(list);
	kmod_unref(ctx);
	return ret;
}

static int unload_module(const char *name)
{
	struct kmod_ctx *ctx;
	struct kmod_module *mod;
	const char *config = NULL;
	struct kmod_list *l, *list = NULL;
	int ret = 0;
	int n = 0;

	ctx = kmod_new(NULL, &config);
	if (!ctx)
		return -1;

	ret = kmod_module_new_from_lookup(ctx, name, &list);
	if (ret < 0) {
		_E("Module %s lookup error", name);
		goto out;
	}

	kmod_list_foreach(l, list) {
		mod = kmod_module_get_module(l);
		if (!mod) {
			_E("Module %s unload error", name);
			ret = -1;
			goto out;
		}

		ret = kmod_module_get_initstate(mod);
		if (ret < 0)
			goto out; /* not loaded */

		ret = kmod_module_remove_module(mod, 0);
		if (ret < 0) {
			_E("Module %s remove error", name);
			goto out;
		}

		++n;
	}

	if (n == 0) {
		_I("Module %s not found", name);
		ret = -1;
		goto out;
	}

	_I("Module %s unloaded\n", name);
	ret = 0;

out:
	kmod_module_unref_list(list);
	kmod_unref(ctx);
	return ret;
}

static int load_gadget()
{
	usbg_state *s;
	int ret = 0;
	FILE *fp = NULL;

	ret = usbg_init(CONFIGFS_PATH, &s);
	if (ret < 0) {
		_E("could not init libusbg");
		return ret;
	}

	fp = fopen(GADGET_SCHEME_PATH, "r");
	if (!fp) {
		_E("could not open gadget scheme");
		ret = -1;
		goto out;
	}

	ret = usbg_import_gadget(s, fp, GADGET_NAME, NULL);
	if (ret < 0) {
		_E("could not import gadget");
		goto out;
	}

out:
	usbg_cleanup(s);
	if (fp)
		fclose(fp);

	return ret;
}

int enable_gadget()
{
	int ret;
	usbg_gadget *g;
	usbg_udc *udc;
	usbg_state *s;

	if (using_legacy)
		return 0;

	ret = usbg_init(CONFIGFS_PATH, &s);
	if (ret < 0)
		return ret;

	g = usbg_get_gadget(s, GADGET_NAME);
	if (!g) {
		_E("could not find gadget");
		ret = -1;
		goto out;
	}

	udc = usbg_get_udc(s, UDC_NAME);
	if (!udc) {
		_E("could not find udc");
		ret = -1;
		goto out;
	}

	ret = usbg_enable_gadget(g, udc);

out:
	usbg_cleanup(s);
	return ret;
}

static void service_started_handler(GDBusConnection  *conn,
	const gchar      *sender,
	const gchar      *path,
	const gchar      *iface,
	const gchar      *name,
	GVariant         *param,
	gpointer          data)
{
	uint32_t id;
	char *unit;
	int ret;

	if (!dh_get_param_from_var(param, "(uoss)", &id, NULL, &unit, NULL)) {
		_E("failed to get params grom gvariant. expected:%s, type:%s", "uoss", g_variant_get_type_string(param));
		return ;
	}

	if (strcmp(unit, SYSTEMD_UNIT_NAME) == 0) {
		ret = enable_gadget();
		if (ret < 0)
			_E("Could not enable gadget");
		else
			_I("Start");

		if (sigid_start) {
			unsubscribe_dbus_signal(NULL, sigid_start);
			sigid_start = 0;
		}
		/*unregister_dbus_signal_handler(SYSTEMD_DBUS_PATH,
				JOB_MANAGER_INTERFACE, JOB_REMOVED);*/
	}
	g_free(unit);
}

static void service_stopped_handler(GDBusConnection  *conn,
	const gchar      *sender,
	const gchar      *path,
	const gchar      *iface,
	const gchar      *name,
	GVariant         *param,
	gpointer          data)

{
	uint32_t id;
	char *unit;
	int ret;
	usbg_state *s;
	usbg_gadget *g;

	if (!dh_get_param_from_var(param, "(uoss)", &id, NULL, &unit, NULL)) {
		_E("failed to get params grom gvariant. expected:%s, type:%s", "uoss", g_variant_get_type_string(param));
		return ;
	}

	if (strcmp(unit, SYSTEMD_SERVICE_NAME) == 0) {
		ret = umount(FFS_PATH);
		if (ret < 0) {
			_E("could not umount functionfs");
			goto out1;
		}

		if (!using_legacy) {
			ret = usbg_init(CONFIGFS_PATH, &s);
			if (ret < 0) {
				_E("could not init libusbg");
				goto out1;
			}

			g = usbg_get_gadget(s, GADGET_NAME);
			if (!g) {
				_E("could not find gadget");
				ret = -1;
				goto out;
			}

			ret = usbg_rm_gadget(g, USBG_RM_RECURSE);
			if (ret < 0) {
				_E("could not remove gadget");
				goto out;
			}
		}

		ret = unload_module("dummy_hcd");
		if (ret < 0) {
			_E("Error unloading module: %d", ret);
			goto out1;
		}

		if (using_legacy) {
			ret = unload_module("g_ffs");
			if (ret < 0) {
				_E("Error unloading module: %d", ret);
				goto out1;
			}
		} else {
			ret = unload_module("usb_f_fs");
			if (ret < 0) {
				_E("Error unloading module: %d", ret);
				goto out;
			}
		}


		_I("stop");

out:
		if (!using_legacy)
			usbg_cleanup(s);
out1:
		if (sigid_stop) {
			unsubscribe_dbus_signal(NULL, sigid_stop);
			sigid_stop = 0;
		}
	}
	g_free(unit);
}

int start()
{
	struct stat st;
	int ret;

	ret = load_module("dummy_hcd", NULL);
	if (ret < 0) {
		_E("Error loading module: %d", ret);
		return ret;
	}

	/* If you need to setup vid, pid or other data use module parameters */
	ret = load_module("g_ffs",
			  "idVendor=0x1D6B"
			  " idProduct=0x104"
			  " bcdDevice=0x0001"
			  " iManufacturer=\"Foo Inc.\""
			  " iProduct=\"Bar Gadget\""
			  " iSerialNumber=\"0123456789\""
			  " generic_ffs_config_str=\"usb host API test config\""
			  " bmAttrs_overwrite=\"0x80\"");
	if (ret < 0) {
		using_legacy = 0;
		_I("Legacy gadget g_ffs not found. Using ConfigFS.");

		ret = load_module("usb_f_fs", NULL);
		if (ret < 0) {
			_E("Error loading module: %d", ret);
			return ret;
		}

		ret = load_gadget();
		if (ret < 0) {
			_E("Error loading gadget: %d", ret);
			goto out;
		}
	} else {
		using_legacy = 1;
	}

	/* TODO make it recusrsive? */
	if (stat(FFS_PATH, &st) < 0) {
		ret = mkdir(FFS_PATH, S_IRWXU | S_IRWXG | S_IROTH);
		if (ret < 0) {
			_E("Error creating ffs directory");
			goto out;
		}
	}

	ret = mount(FFS_INSTANCE_NAME, FFS_PATH, "functionfs", 0, NULL);
	if (ret < 0) {
		_E("Error mounting ffs");
		goto out;
	}

	ret = subscribe_dbus_signal(NULL, SYSTEMD_DBUS_PATH,
		JOB_MANAGER_INTERFACE, JOB_REMOVED,
		service_started_handler, NULL, NULL);
	if (ret <= 0)
		_E("fail to init dbus signal(%d)", ret);
	sigid_start = ret;

	ret = deviced_systemd_start_unit(SYSTEMD_UNIT_NAME);
	if (ret < 0) {
		_E("Error starting daemon");
		/*unregister_dbus_signal_handler(SYSTEMD_DBUS_PATH,
				JOB_MANAGER_INTERFACE, JOB_REMOVED);*/
		if (sigid_start) {
			unsubscribe_dbus_signal(NULL, sigid_start);
			sigid_start = 0;
		}
		goto out;
	}

	status = DEVICE_OPS_STATUS_START;
	return 0;

out:

/* It's already a crash, thus we ignore return values here */

	unload_module("dummy_hcd");
	unload_module("usb_f_fs");
	if (using_legacy)
		unload_module("g_ffs");

	/*unregister_dbus_signal_handler(SYSTEMD_DBUS_PATH,
			JOB_MANAGER_INTERFACE, JOB_REMOVED);*/
	if (sigid_start)
		unsubscribe_dbus_signal(NULL, sigid_start);

	return ret;
}

static int stop()
{
	int ret = 0;

	ret = subscribe_dbus_signal(NULL, SYSTEMD_DBUS_PATH,
		JOB_MANAGER_INTERFACE, JOB_REMOVED,
		service_stopped_handler, NULL, NULL);
	if (ret <= 0) {
		_E("fail to init dbus signal(%d)", ret);
		return ret;
	}

	sigid_stop = ret;

	ret = deviced_systemd_stop_unit(SYSTEMD_UNIT_NAME);
	if (ret < 0) {
		_E("could not stop socket unit");
		goto out;
	}

	ret = deviced_systemd_stop_unit(SYSTEMD_SERVICE_NAME);
	if (ret < 0) {
		_E("could not stop service unit");
		goto out;
	}

	status = DEVICE_OPS_STATUS_STOP;
	return ret;

out:
	/*unregister_dbus_signal_handler(SYSTEMD_DBUS_PATH,
			JOB_MANAGER_INTERFACE, JOB_REMOVED);*/
	if (sigid_stop) {
		unsubscribe_dbus_signal(NULL, sigid_stop);
		sigid_stop = 0;
	}
	return ret;
}

static GVariant *dbus_start(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	static const struct device_ops *usb_host_test_device_ops = NULL;
	int ret;

	if (!usb_host_test_device_ops)
		usb_host_test_device_ops = find_device("usb-host-test");

	ret = usb_host_test_device_ops->start(NORMAL_MODE);
	if (ret != 0 && ret != -EALREADY)
		_E("Failed to start usb-host-test : %d", ret);

	return dbus_handle_new_g_variant_tuple();
}

static GVariant *dbus_stop(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	static const struct device_ops *usb_host_test_device_ops = NULL;
	int ret;

	if (!usb_host_test_device_ops)
		usb_host_test_device_ops = find_device("usb-host-test");

	ret = usb_host_test_device_ops->stop(NORMAL_MODE);
	if (ret != 0 && ret != -EALREADY)
		_E("Failed to stop usb-host-test : %d", ret);

	return dbus_handle_new_g_variant_tuple();
}

static int usb_host_test_start(enum device_flags flags)
{
	if (status == DEVICE_OPS_STATUS_START)
		return -EALREADY;

	return start();
}

static int usb_host_test_stop(enum device_flags flags)
{
	if (status == DEVICE_OPS_STATUS_STOP)
		return -EALREADY;

	return stop();
}

static const dbus_method_s dbus_methods[] = {
	{ "start", NULL, NULL, dbus_start },  /* for devicectl */
	{ "stop",  NULL, NULL, dbus_stop }, /* for devicectl */
	/* Add methods here */
};

static const dbus_interface_u dbus_interface = {
	.oh = NULL,
	.name = DEVICED_INTERFACE_USB_HOST_TEST,
	.methods = dbus_methods,
	.nr_methods = ARRAY_SIZE(dbus_methods),
};

static void usb_host_test_init(void *data)
{
	int ret;

	ret = dbus_handle_add_dbus_object(NULL, DEVICED_PATH_USB_HOST_TEST, &dbus_interface);
	if (ret < 0) {
		_E("Failed to register dbus method! %d", ret);
		return;
	}

	_I("initialized");
}

static void usb_host_test_exit(void *data)
{
	_I("exited");
}

static int usb_host_test_status(void)
{
	return status;
}

static const struct device_ops usb_host_test_device_ops = {
	.name	= "usb-host-test",
	.init	= usb_host_test_init,
	.exit	= usb_host_test_exit,
	.start	= usb_host_test_start,
	.stop	= usb_host_test_stop,
	.status = usb_host_test_status,
};

DEVICE_OPS_REGISTER(&usb_host_test_device_ops)
